#ifndef DLOD_OBJECT_
#define DLOD_OBJECT_

#include <stdio.h>
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include "GraphicalObject.h"

#define MY_INFINITY		1.0e308

/**
 * \brief An object with different levels of detail.
 *
 * This class implements a discrete level of detail object. Several objects that represent different levels
 * of detail may be added to this object along with a distance threshold. Whenever the object is drawn,
 * the appropriate level of detail is chosen by evaluating the distance of the object from the camera.
 */
class DLODObject: public GraphicalObject
{
public:
	/** 
	 * Constructor: Brings this object into a consistent state.
	 *
	 * \param best_level			The default object with most detail.
	 */
	DLODObject (RawObject * best_level);

	/**
	 * Destructor.
	 */
	~DLODObject ();

	/**
	 * Implementation of the draw routine. Chooses the appropriate object to draw depending on the distance between
	 * the camera and the origin of this object.
	 *
	 * \param time					Time at which the object shall be drawn.
	 * \param camera				Camera which captures the object.
	 * \param monitor				Camera which monitors the whole scene, is equal to NULL if we are not in monitoring 
	 *								mode. 
	 * \param show_bounding_volume	States if the bounding volume used for the view frustum culling shall be drawn.
	 * \param culling_mode			Determines the mode used for the view frustum culling (C_NONE, C_BOUNDING_BOX, C_SPHERE
	 *								or C_MIXED).
	 * \param level_of_detail		States wheather level of detail is enabled or not.				
	 */
	void Draw (int time, Camera * camera, Camera * monitor, bool show_bounding_volume, int culling_mode, bool level_of_detail);

	/**
	 * Prints information about the object to the console.
	 */
	void PrintToConsole ();

	/**
	 * Adds a new level of detail to this object. The newly added object will be displayed as soon as its distance
	 * from the camera is bigger than a certain threshold.
	 *
	 * \param next_level			Next level of detail for the object (should be of smaller complexity).
	 * \param distance				The newly added object will be displayed as soon as it is farther away from the
	 *								camera as this threshold.
	 */
	void AddLevel (RawObject * next_level, float distance);

	/**
	 * Returns a new discrete level of detail object that consists of the exact same objects and distance thresholds.
	 *
	 * \return						A clone of this object.
	 */
	DLODObject * GetClone ();

protected:

	/** 
	 * Implementation of the bounding sphere computation. The bounding radius is defined as the
	 * bounding radius of the object which has been used with the constructor.
	 */
	void ComputeBoundingRadius ();

	/** 
	 * Implementation of the bounding box computation. The bounding box is defined as the bounding box
	 * of the object which has been used with the constructor.
	 */
	void ComputeBoundingBox ();


private:
	/** Hide the default constructor.*/
	DLODObject ();

	/** Maximum number of levels. */
	static const int MAX_LEVEL = 8;

	/** The number of different levels that this object has. */
	int num_levels;

	/** Array containing all the objects of the different levels. */
	RawObject * raw_obj[MAX_LEVEL];

	/** Array containing all the distances used to determine the level at which the object shall be drawn. */
	float distances[MAX_LEVEL+1];
};

#endif