#include "Geometry.h"

Point Geometry::PlaneIntersection (const Plane& p1, const Plane& p2, const Plane& p3)
{
	Point result;
	float denominator;
	Vector t1, t2, t3;

	// calculate denominator
	denominator = ScalarProduct (p1.n, CrossProduct (p2.n, p3.n));

	if (denominator == 0.0f)
	{
		result.x = 0.0f;
		result.y = 0.0f;
		result.z = 0.0f;
	}
	else
	{
		t1 = CrossProduct (p2.n, p3.n);
		t2 = CrossProduct (p3.n, p1.n);
		t3 = CrossProduct (p1.n, p2.n);

		result.x = (- t1.x * p1.d - t2.x * p2.d - t3.x * p3.d) / denominator;
		result.y = (- t1.y * p1.d - t2.y * p2.d - t3.y * p3.d) / denominator;
		result.z = (- t1.z * p1.d - t2.z * p2.d - t3.z * p3.d) / denominator;
	}

	return result;
}

Vector Geometry::CrossProduct (const Vector& v1, const Vector& v2)
{
	Vector result;
	
	result.x = v1.y * v2.z - v2.y * v1.z;
	result.y = v1.z * v2.x - v2.z * v1.x;
	result.z = v1.x * v2.y - v2.x * v1.y;

	return result;
}

float Geometry::ScalarProduct (const Vector& v1, const Vector& v2)
{
	return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z;
}

void Geometry::InvertHomogenousMatrix (double * src, double * dest)
{
	dest[0] = src[0];
	dest[1] = src[4];
	dest[2] = src[8];
	dest[3] = 0.0;
	
	dest[4] = src[1];
	dest[5] = src[5];
	dest[6] = src[9];
	dest[7] = 0.0;
	
	dest[8] = src[2];
	dest[9] = src[6];
	dest[10]= src[10];
	dest[11]= 0.0;

	dest[12] = - (dest[0] * src[12] + dest[4] * src[13] + dest[8] * src[14]);
	dest[13] = - (dest[1] * src[12] + dest[5] * src[13] + dest[9] * src[14]);
	dest[14] = - (dest[2] * src[12] + dest[6] * src[13] + dest[10] * src[14]);
	dest[15] = 1.0;
}

void Geometry::InvertHomogenousMatrix (float * src, float * dest)
{
	dest[0] = src[0];
	dest[1] = src[4];
	dest[2] = src[8];
	dest[3] = 0.0;
	
	dest[4] = src[1];
	dest[5] = src[5];
	dest[6] = src[9];
	dest[7] = 0.0;
	
	dest[8] = src[2];
	dest[9] = src[6];
	dest[10]= src[10];
	dest[11]= 0.0;

	dest[12] = - (dest[0] * src[12] + dest[4] * src[13] + dest[8] * src[14]);
	dest[13] = - (dest[1] * src[12] + dest[5] * src[13] + dest[9] * src[14]);
	dest[14] = - (dest[2] * src[12] + dest[6] * src[13] + dest[10] * src[14]);
	dest[15] = 1.0;
}

void Geometry::MatrixMultiply (double * m1, double * m2, double * dest)
{
	dest[0] = m2[0] * m1[0] + m2[1] * m1[4] + m2[2] * m1[8] + m2[3] * m1[12];
	dest[1] = m2[0] * m1[1] + m2[1] * m1[5] + m2[2] * m1[9] + m2[3] * m1[13];
	dest[2] = m2[0] * m1[2] + m2[1] * m1[6] + m2[2] * m1[10] + m2[3] * m1[14];
	dest[3] = m2[0] * m1[3] + m2[1] * m1[7] + m2[2] * m1[11] + m2[3] * m1[15];
	dest[4] = m2[4] * m1[0] + m2[5] * m1[4] + m2[6] * m1[8] + m2[7] * m1[12];
	dest[5] = m2[4] * m1[1] + m2[5] * m1[5] + m2[6] * m1[9] + m2[7] * m1[13];
	dest[6] = m2[4] * m1[2] + m2[5] * m1[6] + m2[6] * m1[10] + m2[7] * m1[14];
	dest[7] = m2[4] * m1[3] + m2[5] * m1[7] + m2[6] * m1[11] + m2[7] * m1[15];
	dest[8] = m2[8] * m1[0] + m2[9] * m1[4] + m2[10] * m1[8] + m2[11] * m1[12];
	dest[9] = m2[8] * m1[1] + m2[9] * m1[5] + m2[10] * m1[9] + m2[11] * m1[13];
	dest[10] = m2[8] * m1[2] + m2[9] * m1[6] + m2[10] * m1[10] + m2[11] * m1[14];
	dest[11] = m2[8] * m1[3] + m2[9] * m1[7] + m2[10] * m1[11] + m2[11] * m1[15];
	dest[12] = m2[12] * m1[0] + m2[13] * m1[4] + m2[14] * m1[8] + m2[15] * m1[12];
	dest[13] = m2[12] * m1[1] + m2[13] * m1[5] + m2[14] * m1[9] + m2[15] * m1[13];
	dest[14] = m2[12] * m1[2] + m2[13] * m1[6] + m2[14] * m1[10] + m2[15] * m1[14];
	dest[15] = m2[12] * m1[3] + m2[13] * m1[7] + m2[14] * m1[11] + m2[15] * m1[15];
}

void Geometry::VectorMultiply (float * m, const Vector & v, Vector & r)
{
	r.x = m[0] * v.x + m[4] * v.y + m[8] * v.z + m[12];
	r.y = m[1] * v.x + m[5] * v.y + m[9] * v.z + m[13];
	r.z = m[2] * v.x + m[6] * v.y + m[10] * v.z + m[14];
}

void Geometry::VectorNormalize (Vector & vect)
{
	float length;

	length = sqrt (vect.x * vect.x + vect.y * vect.y + vect.z * vect.z);
	if (length > 0)
	{
		vect.x /= length;
		vect.y /= length;
		vect.z /= length;
	}
}

float Geometry::VectorLength (Vector & vect)
{
	return sqrt (vect.x * vect.x + vect.y * vect.y + vect.z * vect.z);
}

float Geometry::EnclosingAngle (Vector & v1, Vector & v2)
{
	float cos_of_angle;
	float angle;

	cos_of_angle = ScalarProduct (v1, v2) / (VectorLength (v1) * VectorLength (v2)); 
	angle = acos (cos_of_angle);

	return angle;
}

