#ifndef POSITION_PATH_
#define POSITION_PATH_

#include <stdio.h>
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <math.h>

#include "Geometry.h"

#define PP_LINEAR				1
#define PP_SINE					2
#define PP_COSINE				3
#define PP_JUMP					4


/**
 * \brief Uniquely defines a 6DOF position in 3D space.
 *
 * A 6 degrees of freedom position that specifies translation and orientation in 3D space. The position is
 * given by a set of 3 coordinates that determine the translation and three additional rotation angles in degrees.
 * When applying a position in 3D space, the object is first translated in x y and z direction by the respective
 * x y and z values and then rotated following the RPY-convention. This means, that the object is first rotated
 * around its x axis (roll), then around its current y axis (pitch) and finally around its current z axis (yaw).
 *
 * Using such a position to place an object in 3D space may not be as elegant and robust as using quaternions,
 * but it is pretty intuitive when designing scenes.
 */
typedef struct Position6DOF
{
	float x, y, z;
	float roll, pitch, yaw;
}Position;


/**
 * \brief A 6DOF position in 3D space that may vary over time.
 *
 * A PositionPath is a 6DOF position in 3D space that may vary over time. Such a path is defined by checkpoints
 * that are reached at specific points in time. For times that lie between two checkpoints, the positions of 
 * those checkpoints are interpolated. There are several modes of interpolation possible, either linearly PP_LINEAR,
 * using a cosine (which means smooth start) PP_COSINE, using a sine (which means a smooth stop) or jumping
 * directly from position one to position two PP_JUMP. Each value of the position (x, y, z, roll, pitch, yaw) may
 * either be interpolated in the same way or independently of the other values (for example linearly interpolating
 * between two x coordinates and using a cosine to interpolate between the pitch angles). Each interpolation may
 * additionally have a parameter. While the parameter has no meaning for the linear interpolation, it specifies the
 * number of oscillations when using a sine or cosine interpolation. When using the jumpy interpolation, it must
 * lie between 0.0 and 1.0 and determines the time of the position jump as a the fraction of the timespan between
 * the two checkpoints.
 * 
 * Furthermore, each PositionPath may contain several sequences of checkpoints. Between two such sequences, the 
 * object may not be defined if there is a gap between the end resp. start times of the two sequences. 
 */
class PositionPath
{
public:
	/**
	 * Constructor: Initializes the class and sets it to the zero position {0, 0, 0, 0, 0, 0}
	 */
	PositionPath ();

	/**
	 * Destructor: Destroys the path.
	 */
	~PositionPath ();

	/**
	 * Retrieves the position at any given time. The returned position is only valid if it is indeed
	 * defined at the requested time.
	 *
	 * \param time					Specifies the time which the position shall be retrieved at.
	 * \return						Returns a 6-DOF position.
	 */ 
	Position GetPosition (int time);

	/**
	 * Sets the position to constant values, which are always defined.
	 *
	 * \param x						Value of the x coordinate of the constant position.
	 * \param y						Value of the y coordinate of the constant position.
	 * \param z						Value of the z coordinate of the constant position.
	 * \param roll					Value of the roll angle of the constant position.
	 * \param pitch					Value of the pitch angle of the constant position.
	 * \param yaw					Value of the yaw angle of the constant position.
	 */
	void SetConstantPosition (float x, float y, float z, float roll, float pitch, float yaw);

	/**
	 * States if a position is defined at any given time.
	 *
	 * \param time					Specifies the time at which the path is either defined or not.
	 * \return						true, if the path is defined at the given time; false, if it is not.
	 */
	bool IsDefined (int time);

	/**
	 * States if the position has changing rotation values.
	 *
	 * \return						true if the position contains any changing rotation values over time,
	 *								false otherwise.
	 */
	bool HasDynamicRotation ();

	/**
	 * States if the position contains any rotation (values different from 0.0, 0.0, 0.0 for roll, pitch and yaw).
	 *
	 * \return						true if the position contains any rotation values different from {0.0 0.0, 0.0},
	 *								false, otherwise.
	 */
	bool HasStaticRotation ();

	/**
	 * Retrieves a box which contains the minimal and maximal values for each axis that can be achieved over time.
	 * All points of the path lie within or on the boundary of this box.
	 *
	 * \return						A axis aligned bouding box containing the minimal and maximal values for
	 *								each axix that can be achieved over time.
	 */ 
	BoundingBox GetTranslationBox ();

	/**
	 * Used to construct a path. This routine starts a new path. It may only be called as the first element of the
	 * path or if already existing elements have been completed by calling 'CompletePath (...)'.
	 *
	 * \param offset				The path will start 'offset' milliseconds after the last part of the path.
	 * \param x						x coordinate at the beginning of the path.
	 * \param y						y coordinate at the beginning of the path.
	 * \param z						z coordinate at the beginning of the path.
	 * \param roll					roll angle at the beginning of the path.
	 * \param pitch					pitch angle at the beginning of the path.
	 * \param yaw					yaw angle at the beginning of the path.
	 */
	void StartPath (int offset, float x, float y, float z, float roll, float pitch, float yaw);

	/**
	 * Used to construct a path. This routine ends the current path. It may only be called if a path has already
	 * been started with 'StartPath (...)'.
	 *
	 * \param x						x coordinate at the end of the path.
	 * \param y						y coordinate at the end of the path.
	 * \param z						z coordinate at the end of the path.
	 * \param roll					roll angle at the end of the path.
	 * \param pitch					pitch angle at the end of the path.
	 * \param yaw					yaw angle at the end of the path.
	 */
	void CompletePath (float x, float y, float z, float roll, float pitch, float yaw);

	/**
	 * Used to construct a path. This routine adds a checkpoint to the current path. It may only be called between
	 * 'StartPath (...)' and 'CompletePath (...)'.
	 *
	 * \param x						x coordinate at the checkpoint.
	 * \param y						y coordinate at the checkpoint.
	 * \param z						z coordinate at the checkpoint.
	 * \param roll					roll angle at the checkpoint.
	 * \param pitch					pitch angle at the checkpoint.
	 * \param yaw					yaw angle at the checkpoint.
	 */
	void AddCheckpoint (float x, float y, float z, float roll, float pitch, float yaw);

	/**
	 * Used to construct a path. This routine sets the duration between two points along the path. It may only be called
	 * between 'StartPath (...)' and 'CompletePath (...)'.
	 *
	 * \param time					time which shall elapse between two points along the path.
	 */
	void SetDuration (int time);

	/**
	 * Used to construct a path. Specifies the interpolation between two points of the path. All 6 values of the position
	 * are interpolated using the same method.
	 *
	 * \param mode					Sets the mode of interpolation. Available choices are: PP_LINEAR, PP_SINE, PP_COSINE
	 *								or PP_JUMP.
	 * \param parameter				Sets the parameter for the interpolation. Its meaning differs depending on the mode
	 *								(see class description for more information).
	 */
	void SetGlobalTransition (int mode, float parameter);

	/**
	 * Used to construct a path. Specifies the interpolation between two points of the path. All 6 values of the position
	 * are interpolated independently. Each parameter of this routine can be one of the following: PP_LINEAR, PP_SINE,
	 * PP_COSINE or PP_JUMP.
	 *
	 * \param mode_x				Sets the interpolation mode for the x coordinate.
	 * \param mode_y				Sets the interpolation mode for the y coordinate.
	 * \param mode_z				Sets the interpolation mode for the z coordinate.
	 * \param mode_roll				Sets the interpolation mode for the roll angle.
	 * \param mode_pitch			Sets the interpolation mode for the pitch angle.
	 * \param mode_yaw				Sets the interpolation mode for the yaw angle.
	*/
	void SetTransitionModes (int mode_x, int mode_y, int mode_z, int mode_roll, int mode_pitch, int mode_yaw);

	/**
	 * Used to construct a path. Specifies the parameters for the interpolation between two points of the path. There
	 * is one parameter for each of the 6 values of the position. The meanings of the parameters differ depending on
	 * the chosen interpolation modes.
	 *
	 * \param param_x				Parameter used for the interpolation of the x coordinate.
	 * \param param_y				Parameter used for the interpolation of the y coordinate.
	 * \param param_z				Parameter used for the interpolation of the z coordinate.
	 * \param param_roll			Parameter used for the interpolation of the roll angle.
	 * \param param_pitch			Parameter used for the interpolation of the pitch angle.
	 * \param param_yaw				Parameter used for the interpolation of the yaw angle.
	 */
	void SetParameters (float param_x, float param_y, float param_z, float param_roll, float param_pitch, float param_yaw);

	/**
	 * Shifts the whole path by a certain distance
	 *
	 * \param x						Amount to shift along the x axis
	 * \param y						Amount to shift along the y axis
	 * \param z						Amount to shift along the z axis
	 */
	void Shift (float x, float y, float z);

	/**
	 * Prints the path to the console.
	 */
	void PrintToConsole ();

private:
	/** States if the path contains only constant values which are always defined. */
	bool is_constant;

	/** States if the path contains changing rotation values over time. */
	bool has_dynamic_rotation;

	/** States if the path contains rotation values different from {0, 0 ,0}. */
	bool has_static_rotation;

	/** A box defining the minimal and maximal positions achieved over time for each axis. */
	BoundingBox translation_box;

	/** Constant position, only valid if this 'is_constant' equals 'true'. */
	Position constant_position;

	/** This structure holds the required path information. */
	typedef struct PathNode
	{
		int	start_time;						/**< Time when the transition starts. */
		int end_time;						/**< Time when the transition ends. */
		Position start_position;			/**< The position at the beginning of the transition. */
		Position end_position;				/**< The position at the end of the transition. */
		bool global_transition_evaluator;	/**< Is there just one evaluator for all 6 values? */
		int transition_evaluator[6];		/**< Determines the nterpolation mode between the start and end position. */
		float transition_parameter[6];		/**< Parameters for the transition evaluators. */

		PathNode * next;					/**< Pointer to the previouse node. */
		PathNode * previous;				/**< Pointer to the next node. */
	}Path;

	/** First node of the path. */
	Path * first_node;

	/** internal member for the path construction, states if the path is complete or not. */
	bool not_complete;

	/** internal member for the path construction, points to the current node. */
	PathNode * current_node;

	/** internal member for the path access, points to the node which has been accessed most lately. */
	PathNode * node_pointer;

	/**
	 * Sets the 'node_pointer' to the correct node of the path, such that it points to the node with the
	 * biggest starting time that is still smaller than the requested time. If the requested time is smaller
	 * than the starting time of the first node, 'node_pointer' is set to 'NULL'.
	 *
	 * \param time					Specifies the time which the actual node shall be sought for.
	 */
	void SeekTime (int time);

	/**
	 * Destroys the path and frees its ressources.
	 */
	void DeletePath ();

	/**
	 * Initializes the path.
	 */
	void Initialize ();

	/**
	 * Updates the translation box in a conservative manner if a transition mode contains either a SINE or COSINE
	 * interpolation and its parameter is different from 0.0. This is necessary, because such an interpolation
	 * may result in values between -1 and 1 (instead of 0 and 1), which means that the translation box could
	 * be incorrect.
	 *
	 * \param node					The PathNode for which the translation box must be updated conservatively.
	 */
	void ConservativeUpdate (PathNode * node);

	/**
	 * Returns a position by interpolating between the start and end position of a certain node
	 * using its transition evaluator(s) and parameter(s).
	 *
	 * \param node					Specifies the node which shall be evaluated
	 * \param progress				Specifies the progress between the start and end position as a value
	 *								that ranges from 0.0 to 1.0.
	 * \return						The interpolated position between the start and end position.
	 */
	Position EvaluateNode (PathNode * node, float progress);

	/**
	 * Applies an evaluator.
	 *
	 * \param mode					Specifies the mode of the interpolation (PP_LINEAR, PP_SINE, PP_COSINE or PP_JUMP).
	 * \param value					The value which shall be evaluated.
	 * \param parameter				Specifies the value of the parameter that belongs to the current evaluator.
	 * \return						The value returned by applying the evaluator specified by 'mode' to the 
	 *								'value' with the given 'parameter'.
	 */
	float inline EvaluateValue (int mode, float value, float parameter);

};

#endif