#ifndef TERMINAL_OBJECT_
#define TERMINAL_OBJECT_

#include <stdio.h>
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include "GraphicalObject.h"


/**
 * \brief A graphical object that has no children.
 *
 * A TerminalObject is very similar to a leaf node in a scene graph. The TerminalObject has to be 
 * associated with a RawObject during construction. This object will then be drawn whenever the 
 * TerminalObject shall be visualized. And as the name suggests, a TerminalObject will never
 * have any child objects.
 */
class TerminalObject: public GraphicalObject
{
public:
	/** 
	 * Constructor: Brings the object into a consistent state and associates a RawObject with it.
	 *
	 * \param raw_obj						The object associated with this terminal object.
	 */
	TerminalObject (RawObject * raw_obj);

	/**
	 * Destructor.
	 */
	~TerminalObject ();

	/**
	 * Implementation of the Draw routine. Draws the RawObject that is associated with this terminal object.
	 *
	 * \param time					Time at which the object shall be drawn.
	 * \param camera				Camera which captures the object.
	 * \param monitor				Camera which monitors the whole scene, is equal to NULL if we are not in monitoring 
	 *								mode. 
	 * \param show_bounding_volume	States if the bounding volume used for the view frustum culling shall be drawn.
	 * \param culling_mode			Determines the mode used for the view frustum culling (C_NONE, C_BOUNDING_BOX, C_SPHERE
	 *								or C_MIXED).
	 * \param level_of_detail		States wheather level of detail is enabled or not.				
	 */
	void Draw (int time, Camera * camera, Camera * monitor, bool show_bounding_volume, int culling_mode, bool level_of_detail);

	/**
	 * Prints information about the object to the console.
	 */
	void PrintToConsole ();

protected:
	/** 
	 * Implementation of the bounding sphere routine. The bounding sphere is determined by the RawObject that is 
	 * associated with this terminal object.
	 */
	void ComputeBoundingRadius ();

	/** 
	 * Implementation of the bounding box routine. The bounding box is determined by the RawObject that is 
	 * associated with this terminal object.
	 */
	void ComputeBoundingBox ();


private:
	/** Hide the default constructor. */
	TerminalObject ();
};


#endif