#ifndef TEXTURES_
#define TEXTURES_

#include <stdio.h>
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include "PerlinNoise.h"

#define MAX_TEXTURES			16

#define T_SPACE					0
#define T_PERLIN				1
#define T_PERLIN2				2
#define T_PERLIN3				3
#define T_SKY					4

/**
 * \brief The texture repository.
 *
 * A repository that stores all textures. Cannot be instantiated.
 */
class Textures
{
public:
	/**
	 * Loads all the textures and sends them to the memory of the graphics card.
	 */
	static void Load ();

	/**
	 * Returns the OpenGL identifier of a specific texture.
	 *
	 * \param texture_id				The identifier of the texture which shall be retrieved (see #defines).
	 * \return							The OpenGL identifier for the texture specified by 'texture_id'.
	 */
	static GLuint GetId (int texture_id);

	/**
	 * Releases all textures.
	 */
	static void Release ();

private:
	/** Hide default constructor. */
	Textures ();

	/** Internally stores the OpenGL texture names. */
	static GLuint tex_id[MAX_TEXTURES];

	/**
	 * Loads a windows bitmap file from disk (files with indexed colors not supported).
	 *
	 * \param file_name					A character sequence describing the file to load.
	 *									example: "/Images/Test.bmp".
	 * \param texture_id				The OpenGL texture name which the loaded bitmap shall be bound to.
	 * \return							true if this routine succeeded; false if this routine failed.
	 */
	static bool LoadBMP (char * file_name, GLuint texture_id);

	/**
	 * Creates a perlin texture of an arbitrary size.
	 *
	 * \param perlin					Pointer to a perlin noise class, which describes the actual noise.
	 * \param texture_id				The OpenGL texture name which the created perlin noise shall be bound to.
	 */
	static void CreatePerlinTexture (PerlinNoise * perlin,  GLuint texture_id);

	/**
	 * Creates a mipmap texture.
	 *
	 * \param tex_id					The OpenGL texture name to identify the texture.
	 * \param size_x					Horizontal size of the texture to be created. Does not have to be a value 
	 *									of the form: size_x = 2^i.
	 * \param size_y					Vertical size of the texture to be created. Does not have bo be a value
	 *									of the form: size_y = 2^i.
	 * \param raw_data					Pointer to a sequence of conecutive values between 0 and 255. Since a 
	 *									RGB format is required, the sequence exactly has to be of the size:
	 *									size_x * size_y * 3. 
	 */
	static void CreateTexture (GLuint tex_id, int size_x, int size_y, GLubyte * raw_data);
};


#endif