#ifndef TORUS_
#define TORUS_

#include <stdio.h>
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include "VertexArrayObject.h"

/**
 * \brief A simple torus.
 *
 * A simple torus that uses GLUT to draw itself.
 */
class Torus: public RawObject
{
public:
	/**
	 * Constuctor: Brings the object into a consistent state.
	 *
	 * \param inner_radius				The inner radius of the torus.
	 * \param outer_radius				The outer radius of the torus.
	 * \param n_sides					The number of sides per ring.
	 * \param n_rings					The number of rings that make up the torus.
	 */
	Torus (double inner_radius, double outer_radius, int n_sides, int n_rings);

	/**
	 * Destructor.
	 */
	~Torus ();

	/**
	 * Implementation of the Draw routine. Draws a torus using GLUT.
	 */
	void Draw ();

	/**
	 * Implementation of the GetBoundingRadius routine.
	 */
	double GetBoundingRadius ();

	/**
	 * Implementation of the GetBoundingBox routine.
	 */
	BoundingBox GetBoundingBox ();

private:
	/** Hide the default constructor. */
	Torus ();

	/** Internally stores the inner_radius. */
	double inner_radius;

	/** Internally stores the outer_radius. */
	double outer_radius;

	/** Internally stores the number of sides. */
	int n_sides;

	/** Internally stores the number of rings. */
	int n_rings;
};

/**
 * \brief A simple torus.
 *
 * A textured torus using vertex arrays.
 */
class Torus2: public VertexArrayObject
{
public:
	/**
	 * Constuctor: Brings the object into a consistent state.
	 *
	 * \param radius_t					Distance from the center of the torus to the center of a ring.
	 * \param radius_p					The radius of one single ring.
	 * \param n_sides					Number of sides per ring.
	 * \param n_rings					Number of rings that make up the torus.
	 */
	Torus2 (double radius_t, double radius_p, int n_sides, int n_rings);

	/**
	 * Destructor.
	 */
	~Torus2 ();

	/**
	 * Implementation of the GetBoundingRadius routine.
	 */
	double GetBoundingRadius ();

	/**
	 * Implementation of the GetBoundingBox routine.
	 */
	BoundingBox GetBoundingBox ();

	/**
	 * Associates a texture with this torus.
	 */
	void SetTexture (GLuint texture_id);


private:
	/** Hide the default constructor. */
	Torus2 ();

	/**
	 * Constructs the data for the torus (vertex data, index data and normals).
	 */
	void Construct ();

	/** Internally stores the t parameter of the torus. */
	double r_t;

	/** Internally stores the p parameter of the torus. */
	double r_p;

	/** Internally stores the number of sides per ring. */
	int n_sides;

	/** Internally stores the number of rings that make up the torus. */
	int n_rings;

	/** Internal pointer to the vertex data. */
	float * data;

	/** Internal pointer to the index data. */
	GLuint * indices;
};


#endif