#ifndef TORUS_GENERATOR_
#define TORUS_GENERATOR_

#include <stdio.h>

#include "GroupObject.h"
#include "RawObject.h"
#include "TerminalObject.h"
#include "DLODObject.h"


/**
 * \brief Provides the means to generate the coordinates of points on a torus.
 *
 * Class that provides three different ways of creating a torus. The first two build it recursively
 * and group the objects created during the process together. While the first method creates a
 * TerminalObjects at each end of the recursion and sets its position to a point located on the torus,
 * the second method takes a DLODObject and clones it whenever the recursion ends. The third way
 * of creating a torus is a set of static routines which just iteratively return points
 * lying on a torus.
 */
class TorusGenerator
{
public:
	/**
	 * Creates a torus of objects by recursively calling other routines. During the recursion, two objects are
	 * grouped together, while at the end of the recursion, a new terminal object with a specific appearance is
	 * created.
	 *
	 * \param raw_object					The RawObject which shall be used at the end of the recursion to
	 *										create a terminal object.
	 * \param r_t							The t parameter of the torus creation. This is the radius at which
	 *										the rings of the torus are created.
	 * \param r_p							The p parameter of the torus creation. This is the radius of one 
	 *										single ring.
	 * \param rings							The number of rings that shall be created.
	 * \param slices						The number of points on each created ring.
	 */
	static GraphicalObject * CreateTorus (RawObject * raw_object, double r_t, double r_p, int rings, int slices);

	/**
	 * Similar to 'CreateTorus (...)' above, with the only difference that a discrete level of detail object
	 * is cloned whenever the recursion ends.
	 *
	 * \param dlod_object					The discrete level of detail object which shall be cloned whenever
	 *										the recursion ends.
	 * \param r_t							The t parameter of the torus creation. This is the radius at which
	 *										the rings of the torus are created.
	 * \param r_p							The p parameter of the torus creation. This is the radius of one 
	 *										single ring.
	 * \param rings							The number of rings that shall be created.
	 * \param slices						The number of points on each created ring.
	 */
	static GraphicalObject * CreateTorus (DLODObject * dlod_object,  double r_t, double r_p, int rings, int slices);

	/**
	 * States if there is a valid next point that can be calculated with the routine 'GetNextPoint ()'.
	 *
	 * \return								true if there are still points on the torus to be created,
	 *										false otherwise.
	 */
	static bool HasMore ();

	/**
	 * Starts a new torus.
	 *
	 * \param r_t							The t parameter of the torus creation. This is the radius at which
	 *										the rings of the torus are created.
	 * \param r_p							The p parameter of the torus creation. This is the radius of one 
	 *										single ring.
	 * \param rings							The number of rings that shall be created.
	 * \param slices						The number of points on each created ring.
	 */
	static void StartTorus (double r_t, double r_p, int rings, int slices);

	/**
	 * Returns the next point of the torus creation.
	 *
	 * \return								The next point of the torus creation.
	 */
	static Point GetNextPoint ();

private:
	/** static member to store the current state. */
	static int ring_index;
	/** static member to store the current state. */
	static int slice_index;
	/** static member to store the current state. */
	static double r_t;
	/** static member to store the current state. */
	static double r_p;
	/** static member to store the current state. */
	static int rings;
	/** static member to store the current state. */
	static int slices;

	/**
	 * Creates a part of a torus using recursion.
	 *
	 * \param raw_object					The RawObject which shall be used at the end of the recursion to
	 *										create a terminal object.
	 * \param r_t							The t parameter of the torus creation. This is the radius at which
	 *										the rings of the torus are created.
	 * \param r_p							The p parameter of the torus creation. This is the radius of one 
	 *										single ring.
	 * \param min_t							The minimal t value of the part of the torus which shall be created.
	 * \param max_t							The maximal t value of the part of the torus which shall be created.
	 * \param rings							The number of rings that shall be created.
	 * \param slices						The number of points on each created ring.
	 */
	static GraphicalObject * CreateRings (RawObject * raw_object, double r_t, double r_p, double min_t, double max_t, int rings, int slices);

	/**
	 * Creates a part of a ring of the torus using recursion.
	 *
	 * \param raw_object					The RawObject which shall be used at the end of the recursion to
	 *										create a terminal object.
	 * \param r_t							The t parameter of the torus creation. This is the radius at which
	 *										the rings of the torus are created.
	 * \param r_p							The p parameter of the torus creation. This is the radius of one 
	 *										single ring.
	 * \param curr_t						The current t value for this ring.
	 * \param min_p							The minimal p value of the part of the ring which shall be created.
	 * \param max_p							The maximal p value of the part of the ring which shall be created.
	 * \param slices						The number of points to be created during this routine.
	 */
	static GraphicalObject * CreateRingPart (RawObject * raw_object, double r_t, double r_p, double curr_t, double min_p, double max_p, int slices);

	/**
	 * Creates a part of a torus using recursion. This routine is almost the same as the 'CreateRings (...)' routine above,
	 * with the only difference that a discrete level of detail object is cloned whenever the recursion ends.
	 *
	 * \param dlod_object					The discrete level of detail object  which shall be cloned whenever
	 *										the recursion ends.
	 * \param r_t							The t parameter of the torus creation. This is the radius at which
	 *										the rings of the torus are created.
	 * \param r_p							The p parameter of the torus creation. This is the radius of one 
	 *										single ring.
	 * \param min_t							The minimal t value of the part of the torus which shall be created.
	 * \param max_t							The maximal t value of the part of the torus which shall be created.
	 * \param rings							The number of rings that shall be created.
	 * \param slices						The number of points on each created.
	 */
	static GraphicalObject * CreateRings (DLODObject * dlod_object, double r_t, double r_p, double min_t, double max_t, int rings, int slices);
	
	/*
	 * Creates a part of a ring of the torus using recursion. This routine is almost the same as the 'CreateRingPart (...)' 
	 * routine above, with the only difference that a discrete level of detail object is cloned whenever the recursion
	 * ends.
	 *
	 * \param dlod_object					The discrete level of detail object  which shall be cloned whenever
	 *										the recursion ends.
	 * \param r_t							The t parameter of the torus creation. This is the radius at which
	 *										the rings of the torus are created.
	 * \param r_p							The p parameter of the torus creation. This is the radius of one 
	 *										single ring.
	 * \param curr_t						The current t value for this ring.
	 * \param min_p							The minimal p value of the part of the ring which shall be created.
	 * \param max_p							The maximal p value of the part of the ring which shall be created.
	 * \param slices						The number of points to be created during this routine.
	 */
	static GraphicalObject * CreateRingPart (DLODObject * dlod_object, double r_t, double r_p, double curr_t, double min_p, double max_p, int slices);
};

#endif