#ifndef VERTEX_ARRAY_COMPONENT
#define VERTEX_ARRAY_COMPONENT

#include <stdio.h>
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include "Geometry.h"
#include "RawObject.h"

/**
 * \brief Object that uses vertex arrays.
 *
 * A VertexArrayObject is a specialized form of a RawObjects that supports vertex arrays to display itself.
 * Additional functionality includes the automated calculation of the bounding box as well as the bounding
 * sphere of the object from the vertex data, the computation of the normals of the object from the
 * vertex and index data of the object anf finally a spherical texture coordinate generation routine.
 */ 
class VertexArrayObject: public RawObject
{
public:
	/**
	 * Constructor:	Brings the object into a consistent state.
	 */
	VertexArrayObject ();

	/**
	 * Destructor.
	 */
	~VertexArrayObject ();

	/**
	 * Draws the vertex array object.
	 */
	virtual void Draw ();

	/**
	 * Implementation of the GetBoundingRadius routine. Iterates over the vertex data and returns the maximal
	 * distance from the origin.
	 */
	double GetBoundingRadius ();

	/**
	 * Implementation of the GetBoundingBox routine. Iterates over the vertex data and computes the minimal and
	 * maximal distance along each axis.
	 */
	BoundingBox GetBoundingBox ();

	/**
	 * Prints information about this object to the console.
	 */
	void PrintToConsole ();

	/**
	 * Computes the normals of the object from the vertex and index data. Normals are built by averaging 
	 * for each point the normals of all the faces that use that point. These normals are additionally weighted
	 * by the width of the angle of the respecive faces.
	 *
	 * \param clock_wise				States wheather the faces are indexed clockwise or not.
	 */
	void ComputeNormals (bool clock_wise);

	/**
	 * Computes spherical texture coordinates (and the normal data if necessary) for the object.
	 *
	 * \param texture_id				OpenGL texture name which identifies the texture with which the generated
	 *									coordinates shall be used.
	 */
	void ComputeSphericalTextureCoordinates (GLuint texture_id);

	/** If set to 'true', the normals of the VertexArrayObjects will be drawn whenever such an object is drawn. */
	static bool show_normals;

	/** Lenght of the normals to be drawn. */
	static float normal_length;

protected:
	/**
	 * Sets the vertex data. One vertex must consist of 3 consecutive floating point values (x, y, z).
	 * 
	 * \param data						Pointer to an array which contains the values for the vertices.
	 *									Since the values are not copied, the data to which the pointer points
	 *									must stay valid.
	 * \param num_vertices				Number of vertices contained within the array. This means that the 
	 *									array must contain at least 3 * num_vertices elements.
	 */
	void SetVertexData (float * data, int num_vertices);

	/**
	 * Sets the indices for the vertex data that define the polygon faces.
	 *
	 * \param indices					Pointer to an array which contains indices into the vertex data array.
	 *									Since the values are not copied, the data to which the pointer points
	 *									must stay valid.
	 * \param num_faces					Number of faces contained within the array with the indices. Since one 
	 *									face is defined by 3 consecutive indices, that array must contain at least 
	 *									3 * num_faces elements.
	 */
	void SetVertexIndices (GLuint * indices, int num_faces);

	/**
	 * Sets the texture coordinates for the vertex data. One texture coordinate must consist of exaclty 2
	 * consecutive floating point values (u, v).
	 *
	 * \param data						Pointer to an array which contains the values of the texture coordinates.
	 *									Since the values are not copied, the data to which the pointer points
	 *									must stay valid.
	 * \param texture_id				OpenGL texture name that uniquely identifies the texture that shall be
	 *									used with the texture coordinates.
	 */
	void SetTextureCoordinates (float * data, GLuint texture_id);

private:

	/** Internal pointer to the vertex data. */
	float * vertex_pointer;

	/** Number of vertices contained within the vertex data. */
	int	num_vertices;

	/** Internal pointer to the index data. */
	GLuint * index_pointer;

	/** Number of faces contained within the index data. */
	int num_faces;

	/** Internal pointer to the normal data. */
	float * normal_pointer;	

	/** Internal pointer to the texture coordinate data. */
	float * texture_pointer;

	/** Internal storage of the OpenGL texture name which shall be used along the texture coordinates. */
	GLuint texture_id;

	/** Internal flag indicating wheather the bounding volumes have already been computed or not. */
	bool bounding_volumes_computed;

	/** Internal, precomputed bounding box. */
	BoundingBox bounding_box;

	/** Internal, precomputed bounding radius. */
	double		bounding_radius;

	/**
	 * Draws the normals of the object.
	 */
	void Draw_Normals ();

	/**
	 * Computes the bounding box and bounding radius from the vertex data.
	 */
	void ComputeBoundingVolumes ();
};

#endif